pragma Warnings (Off);
pragma Ada_95;
pragma Source_File_Name (ada_main, Spec_File_Name => "b~gprbind.ads");
pragma Source_File_Name (ada_main, Body_File_Name => "b~gprbind.adb");
pragma Suppress (Overflow_Check);

with System.Restrictions;
with Ada.Exceptions;

package body ada_main is

   E074 : Short_Integer; pragma Import (Ada, E074, "system__os_lib_E");
   E011 : Short_Integer; pragma Import (Ada, E011, "ada__exceptions_E");
   E015 : Short_Integer; pragma Import (Ada, E015, "system__soft_links_E");
   E024 : Short_Integer; pragma Import (Ada, E024, "system__exception_table_E");
   E037 : Short_Integer; pragma Import (Ada, E037, "ada__containers_E");
   E069 : Short_Integer; pragma Import (Ada, E069, "ada__io_exceptions_E");
   E053 : Short_Integer; pragma Import (Ada, E053, "ada__strings_E");
   E055 : Short_Integer; pragma Import (Ada, E055, "ada__strings__maps_E");
   E059 : Short_Integer; pragma Import (Ada, E059, "ada__strings__maps__constants_E");
   E042 : Short_Integer; pragma Import (Ada, E042, "interfaces__c_E");
   E025 : Short_Integer; pragma Import (Ada, E025, "system__exceptions_E");
   E087 : Short_Integer; pragma Import (Ada, E087, "system__object_reader_E");
   E048 : Short_Integer; pragma Import (Ada, E048, "system__dwarf_lines_E");
   E017 : Short_Integer; pragma Import (Ada, E017, "system__soft_links__initialize_E");
   E036 : Short_Integer; pragma Import (Ada, E036, "system__traceback__symbolic_E");
   E110 : Short_Integer; pragma Import (Ada, E110, "ada__strings__utf_encoding_E");
   E118 : Short_Integer; pragma Import (Ada, E118, "ada__tags_E");
   E108 : Short_Integer; pragma Import (Ada, E108, "ada__strings__text_buffers_E");
   E182 : Short_Integer; pragma Import (Ada, E182, "gnat_E");
   E191 : Short_Integer; pragma Import (Ada, E191, "interfaces__c__strings_E");
   E106 : Short_Integer; pragma Import (Ada, E106, "ada__streams_E");
   E175 : Short_Integer; pragma Import (Ada, E175, "system__file_control_block_E");
   E128 : Short_Integer; pragma Import (Ada, E128, "system__finalization_root_E");
   E104 : Short_Integer; pragma Import (Ada, E104, "ada__finalization_E");
   E172 : Short_Integer; pragma Import (Ada, E172, "system__file_io_E");
   E255 : Short_Integer; pragma Import (Ada, E255, "ada__streams__stream_io_E");
   E179 : Short_Integer; pragma Import (Ada, E179, "system__storage_pools_E");
   E220 : Short_Integer; pragma Import (Ada, E220, "system__storage_pools__subpools_E");
   E158 : Short_Integer; pragma Import (Ada, E158, "ada__strings__unbounded_E");
   E316 : Short_Integer; pragma Import (Ada, E316, "system__task_info_E");
   E310 : Short_Integer; pragma Import (Ada, E310, "system__task_primitives__operations_E");
   E448 : Short_Integer; pragma Import (Ada, E448, "system__regpat_E");
   E134 : Short_Integer; pragma Import (Ada, E134, "ada__calendar_E");
   E272 : Short_Integer; pragma Import (Ada, E272, "ada__calendar__delays_E");
   E143 : Short_Integer; pragma Import (Ada, E143, "ada__calendar__time_zones_E");
   E354 : Short_Integer; pragma Import (Ada, E354, "ada__real_time_E");
   E181 : Short_Integer; pragma Import (Ada, E181, "ada__text_io_E");
   E413 : Short_Integer; pragma Import (Ada, E413, "ada__text_io__text_streams_E");
   E216 : Short_Integer; pragma Import (Ada, E216, "gnat__byte_order_mark_E");
   E259 : Short_Integer; pragma Import (Ada, E259, "gnat__calendar_E");
   E261 : Short_Integer; pragma Import (Ada, E261, "gnat__calendar__time_io_E");
   E184 : Short_Integer; pragma Import (Ada, E184, "gnat__directory_operations_E");
   E235 : Short_Integer; pragma Import (Ada, E235, "gnat__dynamic_htables_E");
   E362 : Short_Integer; pragma Import (Ada, E362, "gnat__secure_hashes_E");
   E364 : Short_Integer; pragma Import (Ada, E364, "gnat__secure_hashes__md5_E");
   E360 : Short_Integer; pragma Import (Ada, E360, "gnat__md5_E");
   E251 : Short_Integer; pragma Import (Ada, E251, "gnat__string_split_E");
   E237 : Short_Integer; pragma Import (Ada, E237, "system__pool_global_E");
   E443 : Short_Integer; pragma Import (Ada, E443, "gnat__expect_E");
   E265 : Short_Integer; pragma Import (Ada, E265, "gnat__sockets_E");
   E268 : Short_Integer; pragma Import (Ada, E268, "gnat__sockets__poll_E");
   E277 : Short_Integer; pragma Import (Ada, E277, "gnat__sockets__thin_common_E");
   E270 : Short_Integer; pragma Import (Ada, E270, "gnat__sockets__thin_E");
   E177 : Short_Integer; pragma Import (Ada, E177, "system__regexp_E");
   E132 : Short_Integer; pragma Import (Ada, E132, "ada__directories_E");
   E336 : Short_Integer; pragma Import (Ada, E336, "system__tasking__initialization_E");
   E326 : Short_Integer; pragma Import (Ada, E326, "system__tasking__protected_objects_E");
   E332 : Short_Integer; pragma Import (Ada, E332, "system__tasking__protected_objects__entries_E");
   E344 : Short_Integer; pragma Import (Ada, E344, "system__tasking__queuing_E");
   E350 : Short_Integer; pragma Import (Ada, E350, "system__tasking__stages_E");
   E380 : Short_Integer; pragma Import (Ada, E380, "unicode_E");
   E509 : Short_Integer; pragma Import (Ada, E509, "gprexch_E");
   E397 : Short_Integer; pragma Import (Ada, E397, "sax__htable_E");
   E401 : Short_Integer; pragma Import (Ada, E401, "sax__pointers_E");
   E497 : Short_Integer; pragma Import (Ada, E497, "sax__state_machines_E");
   E468 : Short_Integer; pragma Import (Ada, E468, "schema_E");
   E393 : Short_Integer; pragma Import (Ada, E393, "unicode__ccs_E");
   E417 : Short_Integer; pragma Import (Ada, E417, "unicode__ccs__iso_8859_1_E");
   E419 : Short_Integer; pragma Import (Ada, E419, "unicode__ccs__iso_8859_15_E");
   E424 : Short_Integer; pragma Import (Ada, E424, "unicode__ccs__iso_8859_2_E");
   E427 : Short_Integer; pragma Import (Ada, E427, "unicode__ccs__iso_8859_3_E");
   E429 : Short_Integer; pragma Import (Ada, E429, "unicode__ccs__iso_8859_4_E");
   E431 : Short_Integer; pragma Import (Ada, E431, "unicode__ccs__windows_1251_E");
   E436 : Short_Integer; pragma Import (Ada, E436, "unicode__ccs__windows_1252_E");
   E389 : Short_Integer; pragma Import (Ada, E389, "unicode__ces_E");
   E399 : Short_Integer; pragma Import (Ada, E399, "sax__symbols_E");
   E466 : Short_Integer; pragma Import (Ada, E466, "sax__locators_E");
   E464 : Short_Integer; pragma Import (Ada, E464, "sax__exceptions_E");
   E391 : Short_Integer; pragma Import (Ada, E391, "unicode__ces__utf32_E");
   E439 : Short_Integer; pragma Import (Ada, E439, "unicode__ces__basic_8bit_E");
   E441 : Short_Integer; pragma Import (Ada, E441, "unicode__ces__utf16_E");
   E395 : Short_Integer; pragma Import (Ada, E395, "unicode__ces__utf8_E");
   E462 : Short_Integer; pragma Import (Ada, E462, "sax__models_E");
   E460 : Short_Integer; pragma Import (Ada, E460, "sax__attributes_E");
   E403 : Short_Integer; pragma Import (Ada, E403, "sax__utils_E");
   E376 : Short_Integer; pragma Import (Ada, E376, "dom__core_E");
   E480 : Short_Integer; pragma Import (Ada, E480, "schema__date_time_E");
   E486 : Short_Integer; pragma Import (Ada, E486, "schema__decimal_E");
   E478 : Short_Integer; pragma Import (Ada, E478, "schema__simple_types_E");
   E415 : Short_Integer; pragma Import (Ada, E415, "unicode__encodings_E");
   E411 : Short_Integer; pragma Import (Ada, E411, "dom__core__nodes_E");
   E409 : Short_Integer; pragma Import (Ada, E409, "dom__core__attrs_E");
   E472 : Short_Integer; pragma Import (Ada, E472, "dom__core__character_datas_E");
   E405 : Short_Integer; pragma Import (Ada, E405, "dom__core__documents_E");
   E407 : Short_Integer; pragma Import (Ada, E407, "dom__core__elements_E");
   E452 : Short_Integer; pragma Import (Ada, E452, "input_sources_E");
   E454 : Short_Integer; pragma Import (Ada, E454, "input_sources__file_E");
   E458 : Short_Integer; pragma Import (Ada, E458, "input_sources__strings_E");
   E456 : Short_Integer; pragma Import (Ada, E456, "sax__readers_E");
   E495 : Short_Integer; pragma Import (Ada, E495, "schema__validators_E");
   E474 : Short_Integer; pragma Import (Ada, E474, "schema__readers_E");
   E476 : Short_Integer; pragma Import (Ada, E476, "schema__schema_readers_E");
   E499 : Short_Integer; pragma Import (Ada, E499, "schema__validators__xsd_grammar_E");
   E470 : Short_Integer; pragma Import (Ada, E470, "schema__dom_readers_E");
   E187 : Short_Integer; pragma Import (Ada, E187, "gpr_E");
   E193 : Short_Integer; pragma Import (Ada, E193, "gpr__attr_E");
   E294 : Short_Integer; pragma Import (Ada, E294, "gpr__attr__pm_E");
   E197 : Short_Integer; pragma Import (Ada, E197, "gpr__cset_E");
   E204 : Short_Integer; pragma Import (Ada, E204, "gpr__debug_E");
   E195 : Short_Integer; pragma Import (Ada, E195, "gpr__err_E");
   E232 : Short_Integer; pragma Import (Ada, E232, "gpr__ext_E");
   E371 : Short_Integer; pragma Import (Ada, E371, "gpr__knowledge_E");
   E201 : Short_Integer; pragma Import (Ada, E201, "gpr__names_E");
   E224 : Short_Integer; pragma Import (Ada, E224, "gpr__osint_E");
   E242 : Short_Integer; pragma Import (Ada, E242, "gpr__ali_E");
   E199 : Short_Integer; pragma Import (Ada, E199, "gpr__erroutc_E");
   E206 : Short_Integer; pragma Import (Ada, E206, "gpr__output_E");
   E296 : Short_Integer; pragma Import (Ada, E296, "gpr__scans_E");
   E450 : Short_Integer; pragma Import (Ada, E450, "gpr__sdefault_E");
   E211 : Short_Integer; pragma Import (Ada, E211, "gpr__sinput_E");
   E244 : Short_Integer; pragma Import (Ada, E244, "gpr__snames_E");
   E298 : Short_Integer; pragma Import (Ada, E298, "gpr__strt_E");
   E239 : Short_Integer; pragma Import (Ada, E239, "gpr__tempdir_E");
   E253 : Short_Integer; pragma Import (Ada, E253, "gpr__util_E");
   E358 : Short_Integer; pragma Import (Ada, E358, "gpr__compilation_E");
   E249 : Short_Integer; pragma Import (Ada, E249, "gpr__env_E");
   E302 : Short_Integer; pragma Import (Ada, E302, "gpr__jobserver_E");
   E247 : Short_Integer; pragma Import (Ada, E247, "gpr__tree_E");
   E292 : Short_Integer; pragma Import (Ada, E292, "gpr__dect_E");
   E281 : Short_Integer; pragma Import (Ada, E281, "gpr__nmsc_E");
   E290 : Short_Integer; pragma Import (Ada, E290, "gpr__part_E");
   E300 : Short_Integer; pragma Import (Ada, E300, "gpr__proc_E");
   E279 : Short_Integer; pragma Import (Ada, E279, "gpr__conf_E");
   E501 : Short_Integer; pragma Import (Ada, E501, "gpr__version_E");
   E226 : Short_Integer; pragma Import (Ada, E226, "gpr_build_util_E");
   E507 : Short_Integer; pragma Import (Ada, E507, "gpr__script_E");

   Sec_Default_Sized_Stacks : array (1 .. 1) of aliased System.Secondary_Stack.SS_Stack (System.Parameters.Runtime_Default_Sec_Stack_Size);

   Local_Priority_Specific_Dispatching : constant String := "";
   Local_Interrupt_States : constant String := "";

   Is_Elaborated : Boolean := False;

   procedure finalize_library is
   begin
      declare
         procedure F1;
         pragma Import (Ada, F1, "gpr__util__finalize_body");
      begin
         E253 := E253 - 1;
         F1;
      end;
      declare
         procedure F2;
         pragma Import (Ada, F2, "gpr_build_util__finalize_body");
      begin
         E226 := E226 - 1;
         F2;
      end;
      declare
         procedure F3;
         pragma Import (Ada, F3, "gpr_build_util__finalize_spec");
      begin
         F3;
      end;
      declare
         procedure F4;
         pragma Import (Ada, F4, "gpr__conf__finalize_body");
      begin
         E279 := E279 - 1;
         F4;
      end;
      declare
         procedure F5;
         pragma Import (Ada, F5, "gpr__proc__finalize_body");
      begin
         E300 := E300 - 1;
         F5;
      end;
      declare
         procedure F6;
         pragma Import (Ada, F6, "gpr__nmsc__finalize_body");
      begin
         E281 := E281 - 1;
         F6;
      end;
      declare
         procedure F7;
         pragma Import (Ada, F7, "gpr__knowledge__finalize_body");
      begin
         E371 := E371 - 1;
         F7;
      end;
      E302 := E302 - 1;
      declare
         procedure F8;
         pragma Import (Ada, F8, "gpr__jobserver__finalize_spec");
      begin
         F8;
      end;
      E249 := E249 - 1;
      declare
         procedure F9;
         pragma Import (Ada, F9, "gpr__env__finalize_spec");
      begin
         F9;
      end;
      E358 := E358 - 1;
      declare
         procedure F10;
         pragma Import (Ada, F10, "gpr__compilation__finalize_spec");
      begin
         F10;
      end;
      declare
         procedure F11;
         pragma Import (Ada, F11, "gpr__util__finalize_spec");
      begin
         F11;
      end;
      E187 := E187 - 1;
      declare
         procedure F12;
         pragma Import (Ada, F12, "gpr__sinput__finalize_body");
      begin
         E211 := E211 - 1;
         F12;
      end;
      declare
         procedure F13;
         pragma Import (Ada, F13, "gpr__names__finalize_body");
      begin
         E201 := E201 - 1;
         F13;
      end;
      E232 := E232 - 1;
      declare
         procedure F14;
         pragma Import (Ada, F14, "gpr__knowledge__finalize_spec");
      begin
         F14;
      end;
      declare
         procedure F15;
         pragma Import (Ada, F15, "gpr__ext__finalize_spec");
      begin
         F15;
      end;
      declare
         procedure F16;
         pragma Import (Ada, F16, "gpr__finalize_spec");
      begin
         F16;
      end;
      E470 := E470 - 1;
      declare
         procedure F17;
         pragma Import (Ada, F17, "schema__dom_readers__finalize_spec");
      begin
         F17;
      end;
      E495 := E495 - 1;
      E474 := E474 - 1;
      E476 := E476 - 1;
      declare
         procedure F18;
         pragma Import (Ada, F18, "schema__schema_readers__finalize_spec");
      begin
         F18;
      end;
      declare
         procedure F19;
         pragma Import (Ada, F19, "schema__readers__finalize_spec");
      begin
         F19;
      end;
      declare
         procedure F20;
         pragma Import (Ada, F20, "schema__validators__finalize_spec");
      begin
         F20;
      end;
      E456 := E456 - 1;
      declare
         procedure F21;
         pragma Import (Ada, F21, "sax__readers__finalize_spec");
      begin
         F21;
      end;
      E458 := E458 - 1;
      declare
         procedure F22;
         pragma Import (Ada, F22, "input_sources__strings__finalize_spec");
      begin
         F22;
      end;
      E454 := E454 - 1;
      declare
         procedure F23;
         pragma Import (Ada, F23, "input_sources__file__finalize_spec");
      begin
         F23;
      end;
      E452 := E452 - 1;
      declare
         procedure F24;
         pragma Import (Ada, F24, "input_sources__finalize_spec");
      begin
         F24;
      end;
      E376 := E376 - 1;
      declare
         procedure F25;
         pragma Import (Ada, F25, "dom__core__finalize_spec");
      begin
         F25;
      end;
      E403 := E403 - 1;
      declare
         procedure F26;
         pragma Import (Ada, F26, "sax__utils__finalize_spec");
      begin
         F26;
      end;
      E460 := E460 - 1;
      declare
         procedure F27;
         pragma Import (Ada, F27, "sax__attributes__finalize_spec");
      begin
         F27;
      end;
      E464 := E464 - 1;
      declare
         procedure F28;
         pragma Import (Ada, F28, "sax__exceptions__finalize_spec");
      begin
         F28;
      end;
      E399 := E399 - 1;
      declare
         procedure F29;
         pragma Import (Ada, F29, "sax__symbols__finalize_spec");
      begin
         F29;
      end;
      E401 := E401 - 1;
      declare
         procedure F30;
         pragma Import (Ada, F30, "sax__pointers__finalize_spec");
      begin
         F30;
      end;
      E332 := E332 - 1;
      declare
         procedure F31;
         pragma Import (Ada, F31, "system__tasking__protected_objects__entries__finalize_spec");
      begin
         F31;
      end;
      declare
         procedure F32;
         pragma Import (Ada, F32, "ada__directories__finalize_body");
      begin
         E132 := E132 - 1;
         F32;
      end;
      declare
         procedure F33;
         pragma Import (Ada, F33, "ada__directories__finalize_spec");
      begin
         F33;
      end;
      E177 := E177 - 1;
      declare
         procedure F34;
         pragma Import (Ada, F34, "system__regexp__finalize_spec");
      begin
         F34;
      end;
      declare
         procedure F35;
         pragma Import (Ada, F35, "gnat__sockets__finalize_body");
      begin
         E265 := E265 - 1;
         F35;
      end;
      declare
         procedure F36;
         pragma Import (Ada, F36, "gnat__sockets__finalize_spec");
      begin
         F36;
      end;
      E443 := E443 - 1;
      declare
         procedure F37;
         pragma Import (Ada, F37, "gnat__expect__finalize_spec");
      begin
         F37;
      end;
      E237 := E237 - 1;
      declare
         procedure F38;
         pragma Import (Ada, F38, "system__pool_global__finalize_spec");
      begin
         F38;
      end;
      E251 := E251 - 1;
      declare
         procedure F39;
         pragma Import (Ada, F39, "gnat__string_split__finalize_spec");
      begin
         F39;
      end;
      E360 := E360 - 1;
      declare
         procedure F40;
         pragma Import (Ada, F40, "gnat__md5__finalize_spec");
      begin
         F40;
      end;
      E181 := E181 - 1;
      declare
         procedure F41;
         pragma Import (Ada, F41, "ada__text_io__finalize_spec");
      begin
         F41;
      end;
      E158 := E158 - 1;
      declare
         procedure F42;
         pragma Import (Ada, F42, "ada__strings__unbounded__finalize_spec");
      begin
         F42;
      end;
      E220 := E220 - 1;
      declare
         procedure F43;
         pragma Import (Ada, F43, "system__storage_pools__subpools__finalize_spec");
      begin
         F43;
      end;
      E255 := E255 - 1;
      declare
         procedure F44;
         pragma Import (Ada, F44, "ada__streams__stream_io__finalize_spec");
      begin
         F44;
      end;
      declare
         procedure F45;
         pragma Import (Ada, F45, "system__file_io__finalize_body");
      begin
         E172 := E172 - 1;
         F45;
      end;
      declare
         procedure Reraise_Library_Exception_If_Any;
            pragma Import (Ada, Reraise_Library_Exception_If_Any, "__gnat_reraise_library_exception_if_any");
      begin
         Reraise_Library_Exception_If_Any;
      end;
   end finalize_library;

   procedure adafinal is
      procedure s_stalib_adafinal;
      pragma Import (Ada, s_stalib_adafinal, "system__standard_library__adafinal");

      procedure Runtime_Finalize;
      pragma Import (C, Runtime_Finalize, "__gnat_runtime_finalize");

   begin
      if not Is_Elaborated then
         return;
      end if;
      Is_Elaborated := False;
      Runtime_Finalize;
      s_stalib_adafinal;
   end adafinal;

   type No_Param_Proc is access procedure;
   pragma Favor_Top_Level (No_Param_Proc);

   procedure adainit is
      Main_Priority : Integer;
      pragma Import (C, Main_Priority, "__gl_main_priority");
      Time_Slice_Value : Integer;
      pragma Import (C, Time_Slice_Value, "__gl_time_slice_val");
      WC_Encoding : Character;
      pragma Import (C, WC_Encoding, "__gl_wc_encoding");
      Locking_Policy : Character;
      pragma Import (C, Locking_Policy, "__gl_locking_policy");
      Queuing_Policy : Character;
      pragma Import (C, Queuing_Policy, "__gl_queuing_policy");
      Task_Dispatching_Policy : Character;
      pragma Import (C, Task_Dispatching_Policy, "__gl_task_dispatching_policy");
      Priority_Specific_Dispatching : System.Address;
      pragma Import (C, Priority_Specific_Dispatching, "__gl_priority_specific_dispatching");
      Num_Specific_Dispatching : Integer;
      pragma Import (C, Num_Specific_Dispatching, "__gl_num_specific_dispatching");
      Main_CPU : Integer;
      pragma Import (C, Main_CPU, "__gl_main_cpu");
      Interrupt_States : System.Address;
      pragma Import (C, Interrupt_States, "__gl_interrupt_states");
      Num_Interrupt_States : Integer;
      pragma Import (C, Num_Interrupt_States, "__gl_num_interrupt_states");
      Unreserve_All_Interrupts : Integer;
      pragma Import (C, Unreserve_All_Interrupts, "__gl_unreserve_all_interrupts");
      Detect_Blocking : Integer;
      pragma Import (C, Detect_Blocking, "__gl_detect_blocking");
      Default_Stack_Size : Integer;
      pragma Import (C, Default_Stack_Size, "__gl_default_stack_size");
      Default_Secondary_Stack_Size : System.Parameters.Size_Type;
      pragma Import (C, Default_Secondary_Stack_Size, "__gnat_default_ss_size");
      Bind_Env_Addr : System.Address;
      pragma Import (C, Bind_Env_Addr, "__gl_bind_env_addr");
      Interrupts_Default_To_System : Integer;
      pragma Import (C, Interrupts_Default_To_System, "__gl_interrupts_default_to_system");

      procedure Runtime_Initialize (Install_Handler : Integer);
      pragma Import (C, Runtime_Initialize, "__gnat_runtime_initialize");

      procedure Tasking_Runtime_Initialize;
      pragma Import (C, Tasking_Runtime_Initialize, "__gnat_tasking_runtime_initialize");

      Finalize_Library_Objects : No_Param_Proc;
      pragma Import (C, Finalize_Library_Objects, "__gnat_finalize_library_objects");
      Binder_Sec_Stacks_Count : Natural;
      pragma Import (Ada, Binder_Sec_Stacks_Count, "__gnat_binder_ss_count");
      Default_Sized_SS_Pool : System.Address;
      pragma Import (Ada, Default_Sized_SS_Pool, "__gnat_default_ss_pool");

   begin
      if Is_Elaborated then
         return;
      end if;
      Is_Elaborated := True;
      Main_Priority := -1;
      Time_Slice_Value := -1;
      WC_Encoding := 'b';
      Locking_Policy := ' ';
      Queuing_Policy := ' ';
      Task_Dispatching_Policy := ' ';
      System.Restrictions.Run_Time_Restrictions :=
        (Set =>
          (False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, True, False, False, False, False, 
           False, False, False, False, False, False, False, False, 
           False, False, False, False),
         Value => (0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
         Violated =>
          (True, False, False, False, True, True, True, False, 
           True, False, False, True, True, True, True, False, 
           False, False, False, True, False, False, True, True, 
           False, True, True, False, True, True, True, True, 
           False, True, False, False, False, True, True, False, 
           False, True, False, True, False, True, True, False, 
           True, False, True, False, False, True, True, False, 
           True, False, True, False, False, False, False, False, 
           False, True, False, True, True, True, False, False, 
           True, False, True, True, True, False, True, True, 
           False, True, True, True, True, False, False, False, 
           True, False, False, False, False, True, False, False, 
           True, False, True, False),
         Count => (0, 0, 0, 1, 0, 0, 1, 0, 1, 0),
         Unknown => (False, False, False, False, False, False, True, False, True, False));
      Priority_Specific_Dispatching :=
        Local_Priority_Specific_Dispatching'Address;
      Num_Specific_Dispatching := 0;
      Main_CPU := -1;
      Interrupt_States := Local_Interrupt_States'Address;
      Num_Interrupt_States := 0;
      Unreserve_All_Interrupts := 0;
      Detect_Blocking := 0;
      Default_Stack_Size := -1;

      ada_main'Elab_Body;
      Default_Secondary_Stack_Size := System.Parameters.Runtime_Default_Sec_Stack_Size;
      Binder_Sec_Stacks_Count := 1;
      Default_Sized_SS_Pool := Sec_Default_Sized_Stacks'Address;

      Runtime_Initialize (1);
      Tasking_Runtime_Initialize;

      Finalize_Library_Objects := finalize_library'access;

      Ada.Exceptions'Elab_Spec;
      System.Soft_Links'Elab_Spec;
      System.Exception_Table'Elab_Body;
      E024 := E024 + 1;
      Ada.Containers'Elab_Spec;
      E037 := E037 + 1;
      Ada.Io_Exceptions'Elab_Spec;
      E069 := E069 + 1;
      Ada.Strings'Elab_Spec;
      E053 := E053 + 1;
      Ada.Strings.Maps'Elab_Spec;
      E055 := E055 + 1;
      Ada.Strings.Maps.Constants'Elab_Spec;
      E059 := E059 + 1;
      Interfaces.C'Elab_Spec;
      E042 := E042 + 1;
      System.Exceptions'Elab_Spec;
      E025 := E025 + 1;
      System.Object_Reader'Elab_Spec;
      E087 := E087 + 1;
      System.Dwarf_Lines'Elab_Spec;
      E048 := E048 + 1;
      System.Os_Lib'Elab_Body;
      E074 := E074 + 1;
      System.Soft_Links.Initialize'Elab_Body;
      E017 := E017 + 1;
      E015 := E015 + 1;
      System.Traceback.Symbolic'Elab_Body;
      E036 := E036 + 1;
      E011 := E011 + 1;
      Ada.Strings.Utf_Encoding'Elab_Spec;
      E110 := E110 + 1;
      Ada.Tags'Elab_Spec;
      Ada.Tags'Elab_Body;
      E118 := E118 + 1;
      Ada.Strings.Text_Buffers'Elab_Spec;
      E108 := E108 + 1;
      Gnat'Elab_Spec;
      E182 := E182 + 1;
      Interfaces.C.Strings'Elab_Spec;
      E191 := E191 + 1;
      Ada.Streams'Elab_Spec;
      E106 := E106 + 1;
      System.File_Control_Block'Elab_Spec;
      E175 := E175 + 1;
      System.Finalization_Root'Elab_Spec;
      E128 := E128 + 1;
      Ada.Finalization'Elab_Spec;
      E104 := E104 + 1;
      System.File_Io'Elab_Body;
      E172 := E172 + 1;
      Ada.Streams.Stream_Io'Elab_Spec;
      E255 := E255 + 1;
      System.Storage_Pools'Elab_Spec;
      E179 := E179 + 1;
      System.Storage_Pools.Subpools'Elab_Spec;
      E220 := E220 + 1;
      Ada.Strings.Unbounded'Elab_Spec;
      E158 := E158 + 1;
      System.Task_Info'Elab_Spec;
      E316 := E316 + 1;
      System.Task_Primitives.Operations'Elab_Body;
      E310 := E310 + 1;
      System.Regpat'Elab_Spec;
      E448 := E448 + 1;
      Ada.Calendar'Elab_Spec;
      Ada.Calendar'Elab_Body;
      E134 := E134 + 1;
      Ada.Calendar.Delays'Elab_Body;
      E272 := E272 + 1;
      Ada.Calendar.Time_Zones'Elab_Spec;
      E143 := E143 + 1;
      Ada.Real_Time'Elab_Spec;
      Ada.Real_Time'Elab_Body;
      E354 := E354 + 1;
      Ada.Text_Io'Elab_Spec;
      Ada.Text_Io'Elab_Body;
      E181 := E181 + 1;
      Ada.Text_Io.Text_Streams'Elab_Spec;
      E413 := E413 + 1;
      E216 := E216 + 1;
      Gnat.Calendar'Elab_Spec;
      E259 := E259 + 1;
      Gnat.Calendar.Time_Io'Elab_Spec;
      E261 := E261 + 1;
      Gnat.Directory_Operations'Elab_Spec;
      Gnat.Directory_Operations'Elab_Body;
      E184 := E184 + 1;
      E235 := E235 + 1;
      E362 := E362 + 1;
      E364 := E364 + 1;
      Gnat.Md5'Elab_Spec;
      E360 := E360 + 1;
      Gnat.String_Split'Elab_Spec;
      E251 := E251 + 1;
      System.Pool_Global'Elab_Spec;
      E237 := E237 + 1;
      Gnat.Expect'Elab_Spec;
      E443 := E443 + 1;
      Gnat.Sockets'Elab_Spec;
      Gnat.Sockets.Thin_Common'Elab_Spec;
      E277 := E277 + 1;
      E270 := E270 + 1;
      Gnat.Sockets'Elab_Body;
      E265 := E265 + 1;
      E268 := E268 + 1;
      System.Regexp'Elab_Spec;
      E177 := E177 + 1;
      Ada.Directories'Elab_Spec;
      Ada.Directories'Elab_Body;
      E132 := E132 + 1;
      System.Tasking.Initialization'Elab_Body;
      E336 := E336 + 1;
      System.Tasking.Protected_Objects'Elab_Body;
      E326 := E326 + 1;
      System.Tasking.Protected_Objects.Entries'Elab_Spec;
      E332 := E332 + 1;
      System.Tasking.Queuing'Elab_Body;
      E344 := E344 + 1;
      System.Tasking.Stages'Elab_Body;
      E350 := E350 + 1;
      Unicode'Elab_Body;
      E380 := E380 + 1;
      Gprexch'Elab_Body;
      E509 := E509 + 1;
      E397 := E397 + 1;
      Sax.Pointers'Elab_Spec;
      Sax.Pointers'Elab_Body;
      E401 := E401 + 1;
      E497 := E497 + 1;
      Schema'Elab_Spec;
      E468 := E468 + 1;
      Unicode.Ccs'Elab_Spec;
      E393 := E393 + 1;
      E417 := E417 + 1;
      E419 := E419 + 1;
      E424 := E424 + 1;
      E427 := E427 + 1;
      E429 := E429 + 1;
      E431 := E431 + 1;
      E436 := E436 + 1;
      Unicode.Ces'Elab_Spec;
      E389 := E389 + 1;
      Sax.Symbols'Elab_Spec;
      Sax.Symbols'Elab_Body;
      E399 := E399 + 1;
      E466 := E466 + 1;
      Sax.Exceptions'Elab_Spec;
      Sax.Exceptions'Elab_Body;
      E464 := E464 + 1;
      E391 := E391 + 1;
      E439 := E439 + 1;
      E441 := E441 + 1;
      E395 := E395 + 1;
      Sax.Models'Elab_Spec;
      E462 := E462 + 1;
      Sax.Attributes'Elab_Spec;
      Sax.Attributes'Elab_Body;
      E460 := E460 + 1;
      Sax.Utils'Elab_Spec;
      Sax.Utils'Elab_Body;
      E403 := E403 + 1;
      DOM.CORE'ELAB_SPEC;
      E376 := E376 + 1;
      Schema.Date_Time'Elab_Spec;
      E480 := E480 + 1;
      E486 := E486 + 1;
      Schema.Simple_Types'Elab_Spec;
      E478 := E478 + 1;
      E415 := E415 + 1;
      E411 := E411 + 1;
      E409 := E409 + 1;
      E472 := E472 + 1;
      E407 := E407 + 1;
      E405 := E405 + 1;
      Input_Sources'Elab_Spec;
      Input_Sources'Elab_Body;
      E452 := E452 + 1;
      Input_Sources.File'Elab_Spec;
      Input_Sources.File'Elab_Body;
      E454 := E454 + 1;
      Input_Sources.Strings'Elab_Spec;
      Input_Sources.Strings'Elab_Body;
      E458 := E458 + 1;
      Sax.Readers'Elab_Spec;
      Sax.Readers'Elab_Body;
      E456 := E456 + 1;
      Schema.Validators'Elab_Spec;
      Schema.Readers'Elab_Spec;
      Schema.Schema_Readers'Elab_Spec;
      Schema.Schema_Readers'Elab_Body;
      E476 := E476 + 1;
      Schema.Readers'Elab_Body;
      E474 := E474 + 1;
      E499 := E499 + 1;
      Schema.Validators'Elab_Body;
      E495 := E495 + 1;
      Schema.Dom_Readers'Elab_Spec;
      Schema.Dom_Readers'Elab_Body;
      E470 := E470 + 1;
      GPR'ELAB_SPEC;
      GPR.ATTR'ELAB_SPEC;
      E294 := E294 + 1;
      GPR.CSET'ELAB_BODY;
      E197 := E197 + 1;
      E204 := E204 + 1;
      GPR.EXT'ELAB_SPEC;
      GPR.KNOWLEDGE'ELAB_SPEC;
      GPR.OSINT'ELAB_SPEC;
      GPR.ALI'ELAB_SPEC;
      GPR.ERROUTC'ELAB_SPEC;
      E232 := E232 + 1;
      GPR.OUTPUT'ELAB_BODY;
      E206 := E206 + 1;
      E242 := E242 + 1;
      GPR.NAMES'ELAB_BODY;
      E201 := E201 + 1;
      E296 := E296 + 1;
      GPR.SINPUT'ELAB_SPEC;
      GPR.SINPUT'ELAB_BODY;
      E211 := E211 + 1;
      E199 := E199 + 1;
      E244 := E244 + 1;
      GPR.ATTR'ELAB_BODY;
      E193 := E193 + 1;
      E195 := E195 + 1;
      GPR.TEMPDIR'ELAB_BODY;
      E239 := E239 + 1;
      GPR'ELAB_BODY;
      E187 := E187 + 1;
      GPR.UTIL'ELAB_SPEC;
      GPR.COMPILATION'ELAB_SPEC;
      E358 := E358 + 1;
      GPR.ENV'ELAB_SPEC;
      GPR.ENV'ELAB_BODY;
      E249 := E249 + 1;
      GPR.JOBSERVER'ELAB_SPEC;
      GPR.JOBSERVER'ELAB_BODY;
      E302 := E302 + 1;
      GPR.KNOWLEDGE'ELAB_BODY;
      E371 := E371 + 1;
      GPR.SDEFAULT'ELAB_BODY;
      E450 := E450 + 1;
      GPR.TREE'ELAB_SPEC;
      GPR.TREE'ELAB_BODY;
      E247 := E247 + 1;
      E292 := E292 + 1;
      GPR.NMSC'ELAB_BODY;
      E281 := E281 + 1;
      GPR.PART'ELAB_BODY;
      E290 := E290 + 1;
      GPR.PROC'ELAB_BODY;
      E300 := E300 + 1;
      GPR.CONF'ELAB_SPEC;
      GPR.CONF'ELAB_BODY;
      E279 := E279 + 1;
      GPR.STRT'ELAB_BODY;
      E298 := E298 + 1;
      E501 := E501 + 1;
      Gpr_Build_Util'Elab_Spec;
      Gpr_Build_Util'Elab_Body;
      E226 := E226 + 1;
      GPR.OSINT'ELAB_BODY;
      E224 := E224 + 1;
      GPR.UTIL'ELAB_BODY;
      E253 := E253 + 1;
      E507 := E507 + 1;
   end adainit;

   procedure Ada_Main_Program;
   pragma Import (Ada, Ada_Main_Program, "_ada_gprbind");

   function main
     (argc : Integer;
      argv : System.Address;
      envp : System.Address)
      return Integer
   is
      procedure Initialize (Addr : System.Address);
      pragma Import (C, Initialize, "__gnat_initialize");

      procedure Finalize;
      pragma Import (C, Finalize, "__gnat_finalize");
      SEH : aliased array (1 .. 2) of Integer;

      Ensure_Reference : aliased System.Address := Ada_Main_Program_Name'Address;
      pragma Volatile (Ensure_Reference);

   begin
      if gnat_argc = 0 then
         gnat_argc := argc;
         gnat_argv := argv;
      end if;
      gnat_envp := envp;

      Initialize (SEH'Address);
      adainit;
      Ada_Main_Program;
      adafinal;
      Finalize;
      return (gnat_exit_status);
   end;

--  BEGIN Object file/option list
   --   ./unicode-names.o
   --   ./unicode-names-basic_latin.o
   --   ./unicode.o
   --   ./unicode-names-currency_symbols.o
   --   ./unicode-names-cyrillic.o
   --   ./unicode-names-general_punctuation.o
   --   ./unicode-names-latin_1_supplement.o
   --   ./unicode-names-latin_extended_a.o
   --   ./unicode-names-latin_extended_b.o
   --   ./unicode-names-letterlike_symbols.o
   --   ./unicode-names-spacing_modifier_letters.o
   --   ./dom.o
   --   ./gprexch.o
   --   ./sax.o
   --   ./sax-htable.o
   --   ./sax-pointers.o
   --   ./sax-state_machines.o
   --   ./schema.o
   --   ./unicode-ccs.o
   --   ./unicode-ccs-iso_8859_1.o
   --   ./unicode-ccs-iso_8859_15.o
   --   ./unicode-ccs-iso_8859_2.o
   --   ./unicode-ccs-iso_8859_3.o
   --   ./unicode-ccs-iso_8859_4.o
   --   ./unicode-ccs-windows_1251.o
   --   ./unicode-ccs-windows_1252.o
   --   ./unicode-ces.o
   --   ./sax-symbols.o
   --   ./sax-locators.o
   --   ./sax-exceptions.o
   --   ./unicode-ces-utf32.o
   --   ./unicode-ces-basic_8bit.o
   --   ./unicode-ces-utf16.o
   --   ./unicode-ces-utf8.o
   --   ./sax-encodings.o
   --   ./sax-models.o
   --   ./sax-attributes.o
   --   ./sax-utils.o
   --   ./dom-core.o
   --   ./schema-date_time.o
   --   ./schema-decimal.o
   --   ./schema-simple_types.o
   --   ./unicode-encodings.o
   --   ./dom-core-nodes.o
   --   ./dom-core-attrs.o
   --   ./dom-core-character_datas.o
   --   ./dom-core-elements.o
   --   ./dom-core-documents.o
   --   ./input_sources.o
   --   ./input_sources-file.o
   --   ./input_sources-strings.o
   --   ./sax-readers.o
   --   ./schema-schema_readers.o
   --   ./schema-readers.o
   --   ./schema-validators-xsd_grammar.o
   --   ./schema-validators.o
   --   ./schema-dom_readers.o
   --   ./gpr-attr-pm.o
   --   ./gpr-cset.o
   --   ./gpr-debug.o
   --   ./gpr-opt.o
   --   ./gpr-com.o
   --   ./gpr-ext.o
   --   ./gpr-output.o
   --   ./gpr-ali.o
   --   ./gpr-names.o
   --   ./gpr-scans.o
   --   ./gpr-sinput.o
   --   ./gpr-erroutc.o
   --   ./gpr-snames.o
   --   ./gpr-attr.o
   --   ./gpr-err.o
   --   ./gpr-tempdir.o
   --   ./gpr.o
   --   ./gpr-compilation.o
   --   ./gpr-env.o
   --   ./gpr-jobserver.o
   --   ./gpr-knowledge.o
   --   ./gpr-sdefault.o
   --   ./gpr-tree.o
   --   ./gpr-dect.o
   --   ./gpr-nmsc.o
   --   ./gpr-part.o
   --   ./gpr-proc.o
   --   ./gpr-conf.o
   --   ./gpr-strt.o
   --   ./gpr-version.o
   --   ./gpr_build_util.o
   --   ./gpr-osint.o
   --   ./gpr-util.o
   --   ./gpr-script.o
   --   ./gprbind.o
   --   -L./
   --   -L/build/gprbuild-bootstrap/src/gprbuild-27.0w-20260324-1674D-src/src/
   --   -L/build/gprbuild-bootstrap/src/gprbuild-27.0w-20260324-1674D-src/gpr/src/
   --   -L/build/gprbuild-bootstrap/src/xmlada-27.0w-20260324-16261-src/sax/
   --   -L/build/gprbuild-bootstrap/src/xmlada-27.0w-20260324-16261-src/dom/
   --   -L/build/gprbuild-bootstrap/src/xmlada-27.0w-20260324-16261-src/schema/
   --   -L/build/gprbuild-bootstrap/src/xmlada-27.0w-20260324-16261-src/unicode/
   --   -L/build/gprbuild-bootstrap/src/xmlada-27.0w-20260324-16261-src/input_sources/
   --   -L/usr/lib/gcc/x86_64-pc-linux-gnu/16/adalib/
   --   -static
   --   -lgnarl
   --   -lgnat
   --   -lrt
   --   -lpthread
   --   -ldl
--  END Object file/option list   

end ada_main;
