--  This file is built automatically from data found on the
--  unicode web site (http://www.unicode.org)
--  in version 8.0.0 and thus is a subject to unicode license:
--
--  UNICODE, INC. LICENSE AGREEMENT - DATA FILES AND SOFTWARE
--  See Terms of Use for definitions of Unicode Inc.'s
--  Data Files and Software.
--
--  NOTICE TO USER: Carefully read the following legal agreement.
--  BY DOWNLOADING, INSTALLING, COPYING OR OTHERWISE USING UNICODE INC.'S
--  DATA FILES ("DATA FILES"), AND/OR SOFTWARE ("SOFTWARE"),
--  YOU UNEQUIVOCALLY ACCEPT, AND AGREE TO BE BOUND BY, ALL OF THE
--  TERMS AND CONDITIONS OF THIS AGREEMENT.
--  IF YOU DO NOT AGREE, DO NOT DOWNLOAD, INSTALL, COPY, DISTRIBUTE OR USE
--  THE DATA FILES OR SOFTWARE.
--
--  COPYRIGHT AND PERMISSION NOTICE
--
--  Copyright © 1991-2021 Unicode, Inc. All rights reserved.
--  Distributed under the Terms of Use
--  in https://www.unicode.org/copyright.html.
--
--  Permission is hereby granted, free of charge, to any person obtaining
--  a copy of the Unicode data files and any associated documentation
--  (the "Data Files") or Unicode software and any associated documentation
--  (the "Software") to deal in the Data Files or Software
--  without restriction, including without limitation the rights to use,
--  copy, modify, merge, publish, distribute, and/or sell copies of
--  the Data Files or Software, and to permit persons to whom the Data Files
--  or Software are furnished to do so, provided that either
--  (a) this copyright and permission notice appear with all copies
--  of the Data Files or Software, or
--  (b) this copyright and permission notice appear in associated
--  Documentation.
--
--  THE DATA FILES AND SOFTWARE ARE PROVIDED "AS IS", WITHOUT WARRANTY OF
--  ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
--  WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
--  NONINFRINGEMENT OF THIRD PARTY RIGHTS.
--  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR HOLDERS INCLUDED IN THIS
--  NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL
--  DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
--  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
--  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
--  PERFORMANCE OF THE DATA FILES OR SOFTWARE.
--
--  Except as contained in this notice, the name of a copyright holder
--  shall not be used in advertising or otherwise to promote the sale,
--  use or other dealings in these Data Files or Software without prior
--  written authorization of the copyright holder.
package Unicode.Names.Currency_Symbols is
   pragma Preelaborate;
   pragma Style_Checks (Off);

   Euro_Currency_Sign                      : constant Unicode_Char := 16#20A0#;
   Colon_Sign                              : constant Unicode_Char := 16#20A1#;
   Cruzeiro_Sign                           : constant Unicode_Char := 16#20A2#;
   French_Franc_Sign                       : constant Unicode_Char := 16#20A3#;
   Lira_Sign                               : constant Unicode_Char := 16#20A4#;
   Mill_Sign                               : constant Unicode_Char := 16#20A5#;
   Naira_Sign                              : constant Unicode_Char := 16#20A6#;
   Peseta_Sign                             : constant Unicode_Char := 16#20A7#;
   Rupee_Sign                              : constant Unicode_Char := 16#20A8#;
   Won_Sign                                : constant Unicode_Char := 16#20A9#;
   New_Sheqel_Sign                         : constant Unicode_Char := 16#20AA#;
   Dong_Sign                               : constant Unicode_Char := 16#20AB#;
   Euro_Sign                               : constant Unicode_Char := 16#20AC#;
   Kip_Sign                                : constant Unicode_Char := 16#20AD#;
   Tugrik_Sign                             : constant Unicode_Char := 16#20AE#;
   Drachma_Sign                            : constant Unicode_Char := 16#20AF#;
   German_Penny_Sign                       : constant Unicode_Char := 16#20B0#;
   Peso_Sign                               : constant Unicode_Char := 16#20B1#;
   Guarani_Sign                            : constant Unicode_Char := 16#20B2#;
   Austral_Sign                            : constant Unicode_Char := 16#20B3#;
   Hryvnia_Sign                            : constant Unicode_Char := 16#20B4#;
   Cedi_Sign                               : constant Unicode_Char := 16#20B5#;
   Livre_Tournois_Sign                     : constant Unicode_Char := 16#20B6#;
   Spesmilo_Sign                           : constant Unicode_Char := 16#20B7#;
   Tenge_Sign                              : constant Unicode_Char := 16#20B8#;
   Indian_Rupee_Sign                       : constant Unicode_Char := 16#20B9#;
   Turkish_Lira_Sign                       : constant Unicode_Char := 16#20BA#;
   Nordic_Mark_Sign                        : constant Unicode_Char := 16#20BB#;
   Manat_Sign                              : constant Unicode_Char := 16#20BC#;
   Ruble_Sign                              : constant Unicode_Char := 16#20BD#;
   Lari_Sign                               : constant Unicode_Char := 16#20BE#;
end Unicode.Names.Currency_Symbols;
